unit HogeListView;
(* version 0.0.1  *)
(* Copyright (c) 2001,2002 Twiddle <hetareprog@hotmail.com>  *)

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, ComCtrls;

type
  THogeListView = class(TListView)
  private
    { Private 錾 }
  protected
    { Protected 錾 }
    FObjList: TList;
    FHogeOnData: TLVOwnerDataEvent;
    procedure SetList(AList: TList);
    procedure SetHogeOnData(Handler: TLVOwnerDataEvent);
    procedure HogeOnData(Sender: TObject; Item: TListItem);
  public
    { Public 錾 }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Sort(Compare: TListSortCompare);
    procedure SelectIntoView(Item: TListItem);
    procedure SetTopItem(Item: TListItem);
    property List: TList read FObjList write SetList;
  published
    { Published 錾 }
    property OnData: TLVOwnerDataEvent read FHogeOnData write SetHogeOnData;
  end;

procedure Register;

implementation

type
  TD56TSubItemsCorrector = class (TStringList)
  public
    Dummy: Pointer;
    ToClear: TList;
  end;

procedure Register;
begin
  RegisterComponents('Samples', [THogeListView]);
end;

constructor THogeListView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FObjList := TList.Create;
  OwnerData := False;
end;

destructor THogeListView.Destroy;
begin
  if FObjList <> nil then
    FObjList.Free;
  inherited;
end;

procedure THogeListView.SetList(AList: TList);
begin
  Items.Count := 0;
  if (AList <> nil) and (0 < AList.Count) then
  begin
    FObjList.Assign(AList);
    OwnerData := True;
    TabStop := True;
    Items.Count := FObjList.Count;
  end
  else begin
    OwnerData := False;
    FObjList.Clear;
  end;
end;

procedure THogeListView.Sort(Compare: TListSortCompare);
begin
  FObjList.Sort(Compare);
  Repaint;
end;

procedure THogeListView.SelectIntoView(Item: TListItem);
var
  itemHeight: integer;
  Y: integer;
begin
  Y := Item.Position.Y;
  Selected := item;
  item.Focused := true;
  if Items.Count <= 1 then
    exit;
  itemHeight := (Items[Items.Count -1].Position.Y - Items[0].Position.Y) div Items.Count;
  Scroll(0, Y - TopItem.Position.Y - itemHeight * VisibleRowCount div 2);
end;

procedure THogeListView.SetTopItem(Item: TListItem);
var
  Y: integer;
begin
  Y := Item.Position.Y;
  if Items.Count <= 1 then
    exit;
  Scroll(0, Y - TopItem.Position.Y);
end;

procedure THogeListView.SetHogeOnData(Handler: TLVOwnerDataEvent);
begin
  FHogeOnData := Handler;
  if Assigned(FHogeOnData) then
    inherited OnData := HogeOnData
  else
    inherited OnData := nil;
end;


procedure THogeListView.HogeOnData(Sender: TObject; Item: TListItem);
begin
  TD56TSubItemsCorrector(Item.SubItems).ToClear.Clear;
  if Assigned(FHogeOnData) then
    FHogeOnData(Sender, Item);
end;

end.
