unit UBoardCustomize;

interface

uses
  Classes, IniFiles;

const
  BRDCUST_DEF_MAXRESNUM = -1;

type

  TBoardCustomizeData = class(TObject)
  private
    FRefCount: Integer;
    FMaxRefNum: Integer;
    procedure GetCustomData(Ini: TCustomIniFile; const Section: String);
  public
    constructor Create;
    procedure AddRef;
    procedure Release;
    property MaxResNum: Integer read FMaxRefNum;
  end;

  TBoardCustomize = class(TObject)
  private
    FList: TStringList;
    FNull: TBoardCustomizeData;
    function GetItems(const URL: String): TBoardCustomizeData;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Load(const FileName: String);
    procedure Clear;
    property Items[const URL: String]: TBoardCustomizeData read GetItems;
    property Null: TBoardCustomizeData read FNull;
  end;

implementation

const
  BRDCUST_ID_MAXRESNUM  = 'MaxResNum';

{ TBoardCustomizeData }

constructor TBoardCustomizeData.Create;
begin
  FMaxRefNum := BRDCUST_DEF_MAXRESNUM;
end;

procedure TBoardCustomizeData.AddRef;
begin
  if Self <> nil then
    Inc(FRefCount);
end;

procedure TBoardCustomizeData.Release;
begin
  if Self <> nil then
  begin
    Dec(FRefCount);
    if FRefCount <= 0 then
      Free;
  end;
end;

procedure TBoardCustomizeData.GetCustomData(Ini: TCustomIniFile; const Section: String);
begin
  FMaxRefNum := Ini.ReadInteger(Section, BRDCUST_ID_MAXRESNUM, BRDCUST_DEF_MAXRESNUM);
end;

{ TBoardCustom }

procedure TBoardCustomize.Clear;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    (FList.Objects[i] as TBoardCustomizeData).Release;
  FList.Clear;
end;

constructor TBoardCustomize.Create;
begin
  FNull := TBoardCustomizeData.Create;
  FNull.AddRef;
  FList := TStringList.Create;
end;

destructor TBoardCustomize.Destroy;
begin
  Clear;
  FList.Free;
  FNull.Release;
  inherited;
end;

function TBoardCustomize.GetItems(const URL: String): TBoardCustomizeData;
var
  i: Integer;
begin
  Result := FNull;
  for i := 0 to FList.Count - 1 do
  begin
    if Pos(FList[i], URL) > 0 then
    begin
      Result := TBoardCustomizeData(FList.Objects[i]);
      Break;
    end;
  end;
end;

procedure TBoardCustomize.Load(const FileName: String);
var
  Ini: TMemIniFile;
  Item: TBoardCustomizeData;
  i: Integer;
begin
  Clear;
  Ini := nil;
  try
    Ini := TMemIniFile.Create(FileName);
    try
      Ini.ReadSections(FList);
      for i := 0 to FList.Count - 1 do
      begin
        Item := TBoardCustomizeData.Create;
        try
          Item.GetCustomData(Ini, FList[i]);
          FList.Objects[i] := Item;
          Item.AddRef;
        except
          Item.Free;
        end;
      end;
    except
      Clear;
    end;
  finally
    Ini.Free;
  end;
end;

end.
